/*
  XpressNetMaster.h - library for XpressNet Master protocoll
  Copyright (c) 2016 Philipp Gahtow  All right reserved.

  Version 1.7

  Notice:
  Support for XPressNet Version 3.0 or 3.6!

  Fix:
	- timing issue when change to the next Slot
*/

// ensure this library description is only included once
#ifndef XpressNetMaster_h
#define XpressNetMaster_h

//CONFIG:
#define XNetVersion 0x36
#define XNetID 0x00	//Zentrale: 0x00 = LZ100; 0x01 = LH200; 0x02 = other

// include types & constants of Wiring core API
#if defined(WIRING)
 #include <Wiring.h>
#elif ARDUINO >= 100
 #include <Arduino.h>
#else
 #include <WProgram.h>
#endif

/* From the ATMega datasheet: */
//--------------------------------------------------------------------------------------------
// Which serial port is used, if we have more than one on the chip?
// note that the 328s (the currently produced "smaller" chips) only
// have one serial port, so we force this.
#ifdef __AVR_ATmega328P__
#define SERIAL_PORT_0
#undef SERIAL_PORT_1
#else
//Maybe we are running on a MEGA chip with more than 1 port? If so, you
//can put the serial port to port 1, and use the port 0 for status messages
//to your PC.
#define SERIAL_PORT_1
#undef SERIAL_PORT_0
#endif


// when sending data, do NOT continue until the hardware has sent the data out

#ifdef __AVR_ATmega8__
#define WAIT_FOR_XMIT_COMPLETE {while (!(UCSRA & (1<<TXC))); UCSRA = (1<<TXC); UCSRA = 0;}
#else
#ifdef SERIAL_PORT_0
#define WAIT_FOR_XMIT_COMPLETE {while (!(UCSR0A & (1<<TXC0))); UCSR0A = (1<<TXC0); UCSR0A = 0;}
#else
#define WAIT_FOR_XMIT_COMPLETE {while (!(UCSR1A & (1<<TXC1))); UCSR1A = (1<<TXC1); UCSR1A = 0;}
#endif

#endif

//--------------------------------------------------------------------------------------------

#define XNetTimeUntilNext 2000	//value in microseconds until the next transmission windows starts!
/*An XpressNet device designed to work with XpressNet V3 and later systems must be designed so that it 
begins its transmission within 110 microseconds of receiving its transmission window.  (older X-Bus 
based systems required this transmission to begin with in 40 microseconds.)  Command stations must be 
designed to accept transmissions received up to 120 microseconds after transmitting the window.  The 
difference is to provide a design tolerance between the different types of devices. 
Under normal conditions an XpressNet device must be designed to be able to handle the receipt of its 
next transmission window between 400 microseconds and 500 milliseconds after the receipt of the last window.  */

//Fahrstufen:
#define Loco14 0x00
#define Loco27 0x01
#define Loco28 0x02
#define Loco128 0x04

// XPressnet Call Bytes.
// broadcast to everyone, we save the incoming data and process it later.
#define GENERAL_BROADCAST 0x160
#define FB_BROADCAST 0x1A0

// certain global XPressnet status indicators
#define csNormal 0x00 // Normal Operation Resumed ist eingeschaltet
#define csEmergencyStop 0x01 // Der Nothalt ist eingeschaltet
#define csTrackVoltageOff 0x02 // Die Gleisspannung ist abgeschaltet
#define csShortCircuit 0x04 // Kurzschluss
#define csServiceMode 0x08 // Der Programmiermodus ist aktiv - Service Mode

//XpressNet Befehl, jedes gesendete Byte
#define XNetlength	0		//Lnge
#define XNetheader	1		//Messageheader
#define XNetdata1	2		//Databyte1
#define XNetdata2	3		//Databyte2
#define XNetdata3	4		//Databyte3
#define XNetdata4	5		//Databyte4
#define XNetdata5	6		//Databyte5
#define XNetdata6	7		//Databyte6
#define XNetdata7	8		//Databyte7

// library interface description
class XpressNetMasterClass
{
  // user-accessible "public" interface
  public:
    XpressNetMasterClass(void);	//Constuctor
	void setup(uint8_t FStufen, uint8_t XControl);  //Initialisierung Serial
	void update(void);  			//Set new Data on the Dataline

	void setPower(byte Power);		//Zustand Gleisspannung Melden

	//Control Loco
	void SetLocoBusy(uint8_t UserOps, uint16_t Adr);	//Lok besetzt melden
	void SetLocoInfo(uint8_t UserOps, uint8_t Speed, uint8_t F0, uint8_t F1);	//Lokinfo an XNet Melden
	void SetLocoInfo(uint8_t UserOps, uint8_t Steps, uint8_t Speed, uint8_t F0, uint8_t F1);	//Lokinfo an XNet Melden
	void SetFktStatus(uint8_t UserOps, uint8_t F4, uint8_t F5);	//LokFkt an XNet Melden
	void SetTrntStatus(uint8_t UserOps, uint8_t Address, uint8_t Data); // data=0000 00AA	A=Weichenausgang1+2 (Aktive/Inaktive);

	// public only for easy access by interrupt handlers
	static inline void handle_interrupt();		//Serial Interrupt bearbeiten

  // library-accessible "private" interface
  private:
	  //Variables:
	byte Railpower;		//Data of the actual Power State
	byte Fahrstufe;	//Standard fr Fahrstufe
	byte MAX485_CONTROL; //Port for send or receive control
	uint8_t XNetAdr;	//Adresse des Abzufragenden XNet Device
	unsigned long XSendCount;	//Zeit: Call Byte ausgesendet
	byte XNetMsg[9];	//Serial receive (Length, Header, Data1 to Data7)
	byte rxXOR;			//store the read in XOR

	unsigned int CallByteInquiry;
	unsigned int RequestAck;
	unsigned int DirectedOps;

		//Functions:
	void getNextXNetAdr(void);	//NCHSTE Adr of XNet Device
	void XNetsend(unsigned int *dataString, byte byteCount);	//Sende Datenarray out NOW!
	void getXOR (unsigned int *data, byte length); // calculate the XOR
	unsigned int callByteParity (unsigned int me);	// calculate the parity bit
	//void XNetTransmit (unsigned char data8); //Serial Senden
	void XNetTransmit9 (unsigned int data);
	bool XNetReceive(void);	//Speichern der eingelesenen Daten
	void XNetclear(void);	//Clear a old Message
};

#if defined (__cplusplus)
	extern "C" {
#endif

	extern void notifyXNetPower(uint8_t State) __attribute__((weak));
	extern uint8_t getPowerState() __attribute__((weak));	//give Back Actual Power State
	//Fahrbefehl:
	extern void notifyXNetgiveLocoInfo(uint8_t UserOps, uint16_t Address) __attribute__((weak));
	extern void notifyXNetLocoDrive14(uint16_t Address, uint8_t Speed) __attribute__((weak));
	extern void notifyXNetLocoDrive27(uint16_t Address, uint8_t Speed) __attribute__((weak));
	extern void notifyXNetLocoDrive28(uint16_t Address, uint8_t Speed) __attribute__((weak));
	extern void notifyXNetLocoDrive128(uint16_t Address, uint8_t Speed) __attribute__((weak));
	//Funktionsbefehl:
	extern void notifyXNetgiveLocoFunc(uint8_t UserOps, uint16_t Address) __attribute__((weak));
	extern void notifyXNetLocoFunc1(uint16_t Address, uint8_t Func1) __attribute__((weak));//Gruppe1 0 0 0 F0 F4 F3 F2 F1
	extern void notifyXNetLocoFunc2(uint16_t Address, uint8_t Func2) __attribute__((weak));//Gruppe2 0000 F8 F7 F6 F5
	extern void notifyXNetLocoFunc3(uint16_t Address, uint8_t Func3) __attribute__((weak));//Gruppe3 0000 F12 F11 F10 F9
	extern void notifyXNetLocoFunc4(uint16_t Address, uint8_t Func4) __attribute__((weak));//Gruppe4 F20-F13
	extern void notifyXNetLocoFunc5(uint16_t Address, uint8_t Func5) __attribute__((weak));//Gruppe5 F28-F21
	//Weichenbefehl:
	extern void notifyXNetTrntInfo(uint8_t UserOps, uint8_t Address, uint8_t data) __attribute__((weak));// data=0000 000N	N=Nibble N0-(0,1); N1-(2,3);
	extern void notifyXNetTrnt(uint16_t Address, uint8_t data) __attribute__((weak));// data=0000 000A	A=Weichenausgang (Aktive/Inaktive);
	//CV:
	extern void notifyXNetDirectCV(uint8_t CV, uint8_t data) __attribute__((weak));

#if defined (__cplusplus)
}
#endif


#endif

